import os
import logging


def setup_logging(verbose: bool = False):
    level = logging.DEBUG if verbose else logging.INFO
    logging.basicConfig(level=level, format="%(asctime)s - %(levelname)s - %(message)s")


def ensure_environment_variable(var_name: str):
    value = os.getenv(var_name)
    if not value:
        raise EnvironmentError(f"Environment variable '{var_name}' is not set.")
    return value



#!/usr/bin/env python3

import logging
import os
from diagify.cli import parse_arguments
from diagify.utils import setup_logging
from diagify.llm_client import call_openai,call_openai_user_prompt_only
from diagify.diagram_executor import execute_mingrammer_code
from diagify.validation import identify_incorrect_imports, generate_correction_prompt, create_error_correction_prompt
from diagify.utils import ensure_environment_variable
from diagify.system_prompt import system_prompt


def main():
    args = parse_arguments()
    setup_logging()
    ensure_environment_variable("OPENAI_API_KEY")
    logging.info(f"Diagify is using model: {args.model}")

    # Step 1: Generate Mingrammer code
    logging.info("Generating Mingrammer code...")
    mingrammer_code = call_openai(args.description, system_prompt, args.model)

    # Step 2: Validate the generated code
    from diagify.validation import correct_imports
    incorrect_imports = identify_incorrect_imports(mingrammer_code, correct_imports)

    if incorrect_imports:
        logging.info(f"Incorrect imports found! Diagify will try to correct this...")
        correction_prompt = generate_correction_prompt(mingrammer_code, incorrect_imports, args.description)

        # Step 2a: Correct the imports using OpenAI API
        corrected_code = call_openai_user_prompt_only(correction_prompt, args.model)
        if corrected_code:
            mingrammer_code = corrected_code
        else:
            logging.error("Failed to correct imports. Proceeding with the original code.")
    else:
        logging.info("No issues with imports.")

    # Step 3: Execute the code
    try:
        output_path = execute_mingrammer_code(mingrammer_code)
        if args.output:
            os.rename(output_path, args.output)
            output_path = args.output
        logging.info(f"Final Mingrammer code generated:\n{mingrammer_code}")
        logging.info(f"Success! Diagram saved to {output_path}")
    except Exception as e:

        error_traceback = str(e)
        logging.info(f"Execution failed with error! Diagify will try to correct this...")

        # Step 3a: Generate an error correction prompt
        error_correction_prompt = create_error_correction_prompt(mingrammer_code, error_traceback, args.description)

        # Step 3b: Request OpenAI to fix the error
        fixed_code = call_openai(error_correction_prompt, system_prompt, args.model)

        if fixed_code:
            try:
                output_path = execute_mingrammer_code(fixed_code)
                if args.output:
                    os.rename(output_path, args.output)
                    output_path = args.output
                logging.info(f"Final Mingrammer code generated:\n{mingrammer_code}")
                logging.info(f"Success! Diagram saved to {output_path}")
            except Exception as retry_error:
                logging.error(f"Retry failed with error: {retry_error}")
                exit(1)
        else:
            logging.error("Failed to fix the code after execution error. Exiting.")
            exit(1)


if __name__ == "__main__":
    main()




system_prompt = f'''
You are an expert in the diagram tool Mingrammer and your job is to generate Mingrammer diagram code based on the user's descriptions.  
The response you generate should only contain the python code without anything else.  
It should not even be wrapped in ```python <generated_python_code> ```, instead it should just contain <generated_python_code>.


For example, below is a possible response because it just contains python code: \n

from diagrams import Diagram
from diagrams.aws.compute import Lambda
from diagrams.aws.network import APIGateway
from diagrams.onprem.queue import Kafka

with Diagram("AWS Lambda and Kafka", show=False):
    api_gateway = APIGateway("API Gateway")

    lambda1 = Lambda("Lambda 1")
    lambda2 = Lambda("Lambda 2")
    lambda3 = Lambda("Lambda 3")
    lambda4 = Lambda("Lambda 4")

    kafka_queue = Kafka("Kafka Queue")

    api_gateway >> lambda1 >> kafka_queue
    api_gateway >> lambda2 >> kafka_queue
    api_gateway >> lambda3 >> kafka_queue
    api_gateway >> lambda4 >> kafka_queue
    \n
Notice how in the above example, only python code is generated.

The following are some other examples of Mingrammer diagram generated code.\n
Example 1: Grouped Workers on AWS\n
from diagrams import Diagram
from diagrams.aws.compute import EC2
from diagrams.aws.database import RDS
from diagrams.aws.network import ELB

with Diagram("Grouped Workers", show=False, direction="TB"):
    ELB("lb") >> [EC2("worker1"),
                  EC2("worker2"),
                  EC2("worker3"),
                  EC2("worker4"),
                  EC2("worker5")] >> RDS("events")
\n
Example 2:Clustered Web Services\n

from diagrams import Cluster, Diagram
from diagrams.aws.compute import ECS
from diagrams.aws.database import ElastiCache, RDS
from diagrams.aws.network import ELB
from diagrams.aws.network import Route53

with Diagram("Clustered Web Services", show=False):
    dns = Route53("dns")
    lb = ELB("lb")

    with Cluster("Services"):
        svc_group = [ECS("web1"),
                     ECS("web2"),
                     ECS("web3")]

    with Cluster("DB Cluster"):
        db_primary = RDS("userdb")
        db_primary - [RDS("userdb ro")]

    memcached = ElastiCache("memcached")

    dns >> lb >> svc_group
    svc_group >> db_primary
    svc_group >> memcached\n

Example 3: Event Processing on AWS\n

from diagrams import Cluster, Diagram
from diagrams.aws.compute import ECS, EKS, Lambda
from diagrams.aws.database import Redshift
from diagrams.aws.integration import SQS
from diagrams.aws.storage import S3

with Diagram("Event Processing", show=False):
    source = EKS("k8s source")

    with Cluster("Event Flows"):
        with Cluster("Event Workers"):
            workers = [ECS("worker1"),
                       ECS("worker2"),
                       ECS("worker3")]

        queue = SQS("event queue")

        with Cluster("Processing"):
            handlers = [Lambda("proc1"),
                        Lambda("proc2"),
                        Lambda("proc3")]

    store = S3("events store")
    dw = Redshift("analytics")

    source >> workers >> queue >> handlers
    handlers >> store
    handlers >> dw\n

Example 4: Message Collecting System on GCP\n

from diagrams import Cluster, Diagram
from diagrams.gcp.analytics import BigQuery, Dataflow, PubSub
from diagrams.gcp.compute import AppEngine, Functions
from diagrams.gcp.database import BigTable
from diagrams.gcp.iot import IotCore
from diagrams.gcp.storage import GCS

with Diagram("Message Collecting", show=False):
    pubsub = PubSub("pubsub")

    with Cluster("Source of Data"):
        [IotCore("core1"),
         IotCore("core2"),
         IotCore("core3")] >> pubsub

    with Cluster("Targets"):
        with Cluster("Data Flow"):
            flow = Dataflow("data flow")

        with Cluster("Data Lake"):
            flow >> [BigQuery("bq"),
                     GCS("storage")]

        with Cluster("Event Driven"):
            with Cluster("Processing"):
                flow >> AppEngine("engine") >> BigTable("bigtable")

            with Cluster("Serverless"):
                flow >> Functions("func") >> AppEngine("appengine")

    pubsub >> flow\n

Example 5: Exposed Pod with 3 Replicas on Kubernetes\n

from diagrams import Diagram
from diagrams.k8s.clusterconfig import HPA
from diagrams.k8s.compute import Deployment, Pod, ReplicaSet
from diagrams.k8s.network import Ingress, Service

with Diagram("Exposed Pod with 3 Replicas", show=False):
    net = Ingress("domain.com") >> Service("svc")
    net >> [Pod("pod1"),
            Pod("pod2"),
            Pod("pod3")] << ReplicaSet("rs") << Deployment("dp") << HPA("hpa")\n

Example 6: Stateful Architecture on Kubernetes \n

from diagrams import Cluster, Diagram
from diagrams.k8s.compute import Pod, StatefulSet
from diagrams.k8s.network import Service
from diagrams.k8s.storage import PV, PVC, StorageClass

with Diagram("Stateful Architecture", show=False):
    with Cluster("Apps"):
        svc = Service("svc")
        sts = StatefulSet("sts")

        apps = []
        for _ in range(3):
            pod = Pod("pod")
            pvc = PVC("pvc")
            pod - sts - pvc
            apps.append(svc >> pod >> pvc)

    apps << PV("pv") << StorageClass("sc")\n

Example 7: RabbitMQ Consumers with Custom Nodes\n

from urllib.request import urlretrieve

from diagrams import Cluster, Diagram
from diagrams.aws.database import Aurora
from diagrams.custom import Custom
from diagrams.k8s.compute import Pod

# Download an image to be used into a Custom Node class
rabbitmq_url = "https://jpadilla.github.io/rabbitmqapp/assets/img/icon.png"
rabbitmq_icon = "rabbitmq.png"
urlretrieve(rabbitmq_url, rabbitmq_icon)

with Diagram("Broker Consumers", show=False):
    with Cluster("Consumers"):
        consumers = [
            Pod("worker"),
            Pod("worker"),
            Pod("worker")]

    queue = Custom("Message queue", rabbitmq_icon)

    queue >> consumers >> Aurora("Database")

Also, to remind you of some correct import paths:\n
The import path for dynamoDB is 'from diagrams.aws.database import Dynamodb'\n
remember that this is case sensitive so 'from diagrams.aws.database import DynamoDB' will not work because 'DynamoDB' 
has incorrect casing.\n
Also, the import path for Redis is 'from diagrams.onprem.inmemory import Redis'\n
Never use 'from diagrams.custom import Custom' because we do not have any images to actually use.
Never use ClusterRole as a context manager e.g.\n
with ClusterRole("TPservice"):
        k8s_cluster = Lambda("Kubernetes Cluster")\n

'''


import os
import logging
from openai import OpenAI

def get_openai_client():
    api_key = os.getenv("OPENAI_API_KEY")
    if not api_key:
        raise ValueError("OPENAI_API_KEY environment variable is not set.")
    return OpenAI(api_key=api_key)


def call_openai(description, system_prompt, model):
    client = get_openai_client()
    prompt = f"Generate Mingrammer diagrams code for the following description:\n\n{description}\n\nCode:"
    response = client.chat.completions.create(
        model=model,
        messages=[{"role": "system", "content": system_prompt}, {"role": "user", "content": prompt}],
        max_tokens=1084,
        temperature=0.5,
    )
    return response.choices[0].message.content


def call_openai_user_prompt_only(correction_prompt, model="gpt-4"):
    client = get_openai_client()
    response = client.chat.completions.create(
        model=model,
        messages=[{"role": "user", "content": correction_prompt}],
        max_tokens=1084,
        temperature=0.5
    )
    corrected_code = response.choices[0].message.content.strip()
    return corrected_code



import difflib
import os
import importlib
import diagrams


def populate_exec_globals_and_get_imports() -> list:
    """Populate exec_globals with Mingrammer diagram classes and return valid import statements."""
    import_statements = []
    diagrams_path = os.path.dirname(diagrams.__file__)

    for root, dirs, files in os.walk(diagrams_path):
        for file in files:
            if file.endswith(".py") and file != "__init__.py":
                relative_path = os.path.relpath(os.path.join(root, file), diagrams_path)
                module_name = f"diagrams.{relative_path[:-3].replace(os.path.sep, '.')}"

                try:
                    module = importlib.import_module(module_name)
                    for attr_name in dir(module):
                        attr = getattr(module, attr_name)
                        if isinstance(attr, type):
                            import_statements.append(f"from {module_name} import {attr_name}")
                except ImportError:
                    pass
    return import_statements


# Generate `correct_imports` list dynamically
correct_imports = populate_exec_globals_and_get_imports()


def identify_incorrect_imports(generated_code: str, correct_imports: list) -> dict:
    """
    Identify incorrect import statements in the generated code and provide suggestions
    with emphasis on matching the imported entity (e.g., class or function name).
    """
    incorrect_imports = {}

    for line in generated_code.splitlines():
        # Check for import statements
        if (line.startswith("from diagrams") or line.startswith("import diagrams")) and line not in [
            "from diagrams import Diagram",
            "from diagrams import Cluster, Diagram",
            "from diagrams import Diagram, Cluster"
        ]:
            if line not in correct_imports:
                # Extract the imported entity (e.g., "Kubernetes" in the example)
                imported_entity = line.split("import")[-1].strip() if "import" in line else ""

                # Filter correct_imports to focus on matching the imported entity
                focused_matches = [
                    match for match in correct_imports
                    if imported_entity and imported_entity in match
                ]

                # If no focused matches, use general closest matches
                closest_matches = focused_matches if focused_matches else difflib.get_close_matches(
                    line, correct_imports, n=10, cutoff=0.6
                )

                # Filter out private/special names
                closest_matches = [
                    match for match in closest_matches
                    if not match.split()[-1].startswith("_")
                ]

                if closest_matches:
                    incorrect_imports[line] = closest_matches

    return incorrect_imports


def generate_correction_prompt(generated_code: str, incorrect_imports: dict, original_prompt: str) -> str:
    """Generate a prompt to correct invalid imports."""
    incorrect_imports_text = "\n".join(
        f"- Incorrect: {incorrect} | Suggested: {suggested}"
        for incorrect, suggested in incorrect_imports.items()
    )
    return f"""
    You are an expert code debugger.  I will show you some python code that uses the Mingrammer diagramming tool
    That uses some incorrect imports that I want you to fix.  The original intention of the code was the following:\n
    {original_prompt}\n  Below is the code with the incorrect imports: \n
    {generated_code}\n
    The incorrect import lines and some possible suggestions for correct imports is below.\n
    {incorrect_imports_text}\n
    Please replace only the incorrect import lines with the suggested imports. Keep the rest of the code unchanged.
    The response you generate should only contain the python code without anything else.  
    It should not even be wrapped in ```python <generated_python_code> ```, instead it should just contain <generated_python_code>.
    """


def create_error_correction_prompt(code: str, error: str, description: str) -> str:

    return f"""
    You are a Python expert specializing in diagram generation. I attempted to run the following code, 
    which was intended to generate a diagram based on this description:\n\n{description}\n\n
    However, it threw the following error:\n\n{error}\n\n
    Here is the code that caused the error:\n\n{code}\n\n
    Please fix the code so that it runs without errors and achieves the intended purpose. 
    The response you generate should only contain the python code without anything else.  
    It should not even be wrapped in ```python <generated_python_code> ```, instead it should just contain <generated_python_code>.
    """


import subprocess
import tempfile
import os
import glob
import logging
import sys


def execute_mingrammer_code(code: str) -> str:
    with tempfile.NamedTemporaryFile(delete=False, suffix=".py") as tmp_script:
        script_path = tmp_script.name
        tmp_script.write(code.encode())
        tmp_script.flush()

        try:
            # Capture stdout and stderr from the executed script
            result = subprocess.run(
                [sys.executable, script_path],  # Use the current Python interpreter
                check=True,
                cwd=os.getcwd(),
                text=True,
                capture_output=True,
            )
            logging.info(f"Script executed successfully: {result.stdout}")
            output_path = get_latest_png_file()
            logging.info(f"Diagram successfully generated: {output_path}")
            return output_path
        except subprocess.CalledProcessError as e:
            # Capture and log the detailed error output
            logging.info("Mingrammer code execution failed!")
            raise RuntimeError(f"Failed to execute the generated Mingrammer code. Error:\n{e.stderr}")



def get_latest_png_file(directory=".") -> str:
    list_of_files = glob.glob(f"{directory}/*.png")
    if not list_of_files:
        raise FileNotFoundError("No PNG files found.")
    return max(list_of_files, key=os.path.getctime)



from setuptools import setup, find_packages

setup(
    name="diagify",
    version="0.1.4",
    packages=find_packages(include=["diagify", "diagify.*"]),
    install_requires=[
        "openai",
        "diagrams",
        "graphviz"
    ],
    entry_points={
        "console_scripts": [
            "diagify=diagify.main:main",
        ],
    },
    author="Alex Minnaar",
    author_email="minnaaralex@gmail.com",
    description="A tool to generate diagrams from natural language using Mingrammer.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/alexminnaar/diagify",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)



import argparse


def parse_arguments():
    parser = argparse.ArgumentParser(description="Diagify: Translate natural language into technical diagrams.")
    parser.add_argument("description", type=str, help="The description of the diagram to generate.")
    parser.add_argument(
        "--model",
        type=str,
        default="gpt-4o-mini",
        help="OpenAI model to use for generation (default: gpt-4o-mini)",
        choices=["gpt-4o-mini", "gpt-4o"]
    )
    parser.add_argument(
        "--output",
        type=str,
        default=None,
        help="Optional output path for the generated diagram (default: current directory).",
    )

    return parser.parse_args()
